#define _GNU_SOURCE

#include <assert.h>
#include <errno.h>
#include <getopt.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include "config.h"
#include "opts.h"

#define TIMEOUT_DEFAULT 5
#define TIMEOUT_MAX 3600

/* Initialize global program options.
 */
opts_t opts =
   {
      .connect = NULL,
      .is_attached = false,
      .ntn_name = NULL,
      .pidfile_name = "/tmp/pgear.pid",
      .run = NULL,
      .sql = NULL,
      .timeout_sec = TIMEOUT_DEFAULT,
      .verbosity = 0
   };
/*---------------------------------------------------------------------------*/

/* Print help, in response to the user asking for help or to the user
 * misusing other options, and exit the program.
 */
void
print_usage (FILE* stream, int exit_code)
{
   fprintf(stream, "Usage: %s [OPTION]... { --run=CMD | --sql=SQL } NAME\n",
           PACKAGE_NAME);
   fprintf(stream,
"Listen for notifications of NAME and in response run CMD or run SQL.\n"
"In SQL, use `$1' as the payload parameter, e.g. --sql='SELECT myfunc($1)'.\n\n"
"-a --attached     Remain attached instead of daemonizing.\n"
"-c --connect=NAME Use this PostgreSQL connection configuration service name.\n"
"-p --pidfile=NAME Write pid to file NAME; default %s.\n"
"-t --timeout=N    Wait Ns for startup; 0 means indefinitely; default %d.\n"
"-v --verbose      Print verbose messages.  Repeat for more verbosity.\n"
"   --help         Display this usage information.\n"
"   --version      Print version number.\n\n"
"Report bugs to <" PACKAGE_BUGREPORT ">.  Home page <" PACKAGE_URL ">.\n",
           opts.pidfile_name, TIMEOUT_DEFAULT
   );
   exit(exit_code);
}
/*---------------------------------------------------------------------------*/

/* Print version.
 */
void
print_version (void)
{
   printf(PACKAGE_STRING
"\nCopyright (C) 2022 Nick <barnold@tilde.club>\n"
"License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\n"
"This is free software: you are free to change and redistribute it.\n"
"There is NO WARRANTY, to the extent permitted by law.\n"
   );
   exit(EXIT_SUCCESS);
}
/*---------------------------------------------------------------------------*/

/* Return the timeout seconds in the expected range or die.
 */
unsigned int
get_timeout (const char *restrict optarg)
{
   char *tail;
   long int n = strtol(optarg, &tail, 10);
   /* Demand we got an integer in the range and nothing else. */
   if (n < 0 || TIMEOUT_MAX < n || '\0' != *tail)
   {
      fprintf(stderr, "%s: use a timeout in range 0..%d.\n",
              PACKAGE_NAME, TIMEOUT_MAX);
      print_usage(stderr, EXIT_FAILURE);
   }
   return (unsigned short) n;
}
/*---------------------------------------------------------------------------*/

void
init_opts_from_cmdline (int argc, char **argv)
{

   /* The 1-char or -1 value returned by getopt_long().
    */
   int next_option;

   /* An array describing valid long options. */
   const struct option long_options[] =
      {
         { "attached", 0, NULL, 'a' },
         { "connect",  1, NULL, 'c' },
         { "help",     0, NULL, 'h' },
         { "pidfile",  1, NULL, 'p' },
         { "run",      1, NULL, 'r' },
         { "sql",      1, NULL, 's' },
         { "timeout",  1, NULL, 't' },
         { "verbose",  0, NULL, 'v' },
         { "version",  0, NULL, 'V' },
         { NULL, 0, NULL, 0 } /* Required at end of array. */
      };

   do
   {
      next_option = getopt_long(argc, argv, "ac:hp:r:s:t:vV", long_options, NULL);
      switch (next_option)
      {
      case 'a':
         opts.is_attached = true;
         break;
      case 'c':
         opts.connect = optarg;
         break;
      case 'h':
         print_usage(stdout, EXIT_SUCCESS);
         assert(false);  /* Should be unreachable. */
         abort();
      case 'p':
         opts.pidfile_name = optarg;
         break;
      case 'r':
         opts.run = optarg;
         break;
      case 's':
         opts.sql = optarg;
         break;
      case 't':
         opts.timeout_sec = get_timeout(optarg);
         break;
      case 'v':
         ++opts.verbosity;
         break;
      case 'V':
         print_version();
         assert(false);  /* Should be unreachable. */
         abort();
      case '?':          /* '?' for 'invalid option'. */
         print_usage(stderr, EXIT_FAILURE);
         assert(false);  /* Should be unreachable. */
         abort();
      case -1:           /* Done with options.  */
         break;
      default:
         assert(false);  /* Should be unreachable. */
         abort();
      }
   }
   while (-1 != next_option);

   /* Want exactly one of --sql or --run.
    */
   if ((opts.run && opts.sql) || (!opts.run && !opts.sql))
   {
      print_usage(stderr, EXIT_FAILURE);
   }

   /* Want exactly one non-option argument, the notification name.
    */
   if (1 != argc - optind) print_usage(stderr, EXIT_FAILURE);
   opts.ntn_name = argv[optind];

   return;
}
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
