#define _GNU_SOURCE

#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include "libpq-fe.h"
#include "logr.h"
#include "opts.h"
#include "pger.h"

/* The database connection.
 */
char *conninfo = "";
PGconn *conn;

/* To block SIG{INT,TERM} while a --run=CMD is in progress.
 */
sigset_t run_cmd_mask;

/* For use by the signal handler.
 */
volatile sig_atomic_t shutdown_signalled = 0;

/* Whether the server provides a payload and whether it's wanted
 * by the --sql=SQL.
 */
bool is_pd_avail = false;
bool is_pd_want = false;

/* The function to invoke to respond to a notification.  We'll set it to
 * one of do_response_{cmd,sql} later.
 */
void (*do_response) (const PGnotify *notify) = NULL;
/*---------------------------------------------------------------------------*/

void
catch_signal_to_shutdown (int signum)
{
   /* assert to silence compiler warning. */
   assert(signum > 0);
   shutdown_signalled = 1;
   return;
}
/*---------------------------------------------------------------------------*/

/*
  Run given SQL.
*/
void
do_sql (const char *sql, const PGnotify *notify)
{
   PGresult *res = NULL;
   logg(3, "%s running `%s'...\n", __func__, sql);
   if (is_pd_want)
   {
      logg(3, "%s looking for a payload.\n", __func__);
      const char * param_values[] = { notify ? notify->extra : "" };
      res = PQexecParams(conn, sql, 1, NULL, param_values, NULL, NULL, 0);
   }
   else
   {
      logg(3, "%s no payload.\n", __func__);
      res = PQexecParams(conn, sql, 0, NULL, NULL, NULL, NULL, 0);
   }
   ExecStatusType status = PQresultStatus(res);
   switch (status)
   {
      /* These are 'good', nothing more to do. */
      case PGRES_COMMAND_OK:
      case PGRES_TUPLES_OK:
      case PGRES_COPY_OUT:
      case PGRES_COPY_IN:
         break;
      /* Anything else is 'bad'. */
      default:
         PQclear(res);
         logg_exit("failed `%s': %s / %s\n",
                     sql, PQresStatus(status), PQerrorMessage(conn));
   }
   logg(3, "%s PQresStatus `%s'.\n", __func__, PQresStatus(status));
   PQclear(res);
   return;
}
/*---------------------------------------------------------------------------*/

/*
  Run response SQL.
*/
void
do_response_sql (const PGnotify *notify)
{
   do_sql(opts.sql, notify);
   return;
}
/*---------------------------------------------------------------------------*/

/* Run response command.
 */
void
do_response_cmd (const PGnotify *notify)
{
   char *cmd = NULL;
   logg(3, "%s: notify is %s\n", __func__, notify ? "set" : "null");
   if (is_pd_avail)
   {
      const char *payload = (notify ? notify->extra : NULL);
      if (asprintf(&cmd, "%s \"%s\"", opts.run, payload) < 0)
      {
         logg_exit("failed to allocate cmd string: %s\n", strerror(errno));
      }
      logg(3, "cmd `%s'.\n", cmd);
   }
   else
   {
      cmd = opts.run;
   }
   logg(3, "performing `%s'...\n", cmd);
   /* Before the run, temporarily block SIG{INT,TERM}. */
   if (-1 == sigprocmask(SIG_BLOCK, &run_cmd_mask, NULL))
   {
      logg_exit("failed sigprocmask block, %s line %d: %s\n",
                __FILE__, __LINE__, strerror(errno));
   }
   int sysret = system(cmd);
   if (-1 == sysret) logg_exit("system() failed: %s\n", strerror(errno));
   if (0 != sysret) logg_exit("system() failed, returned %d.\n", sysret);
   /* Run succeeded so unblock. */
   if (-1 == sigprocmask(SIG_UNBLOCK, &run_cmd_mask, NULL))
   {
      logg_exit("failed sigprocmask unblock, %s line %d: %s\n",
                __FILE__, __LINE__, strerror(errno));
   }
   logg(3,"completed `%s'.\n", cmd);
   if (is_pd_avail) free(cmd);
   return;
}
/*---------------------------------------------------------------------------*/

/* Clean up the connection when no longer wanted.
 */
void
finish_conn (void)
{
   logg(2, "%s about to call PQfinish...\n", __func__);
   PQfinish(conn);
   logg(2, "%s called PQfinish.\n", __func__);
   return;
}
/*---------------------------------------------------------------------------*/

void
connect_to_db()
{
   if (NULL != opts.connect)
   {
      if (asprintf(&conninfo, "service=%s", opts.connect) < 0)
      {
         logg_exit("failed to allocate conninfo string: %s\n", strerror(errno));
      }
      logg(3, "from connect option `%s', built conninfo string `%s'.\n",
           opts.connect, conninfo);
   }
   logg(2, "connecting using conninfo string `%s'...\n", conninfo);
   conn = PQconnectdb(conninfo);
   if (CONNECTION_OK != PQstatus(conn))
   {
      logg_exit("failed connection with conninfo `%s': %s\n",
                conninfo, PQerrorMessage(conn));
   }

   logg(3, "setting up handler for signals...\n");
   struct sigaction sa;
   sa.sa_handler = catch_signal_to_shutdown;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = 0;
   if (-1 == sigaction(SIGTERM, &sa, NULL))
   {
      logg_exit("failed sigaction for `%s': %s\n",
                  strsignal(SIGTERM), strerror(errno));
   }
   if (-1 == sigaction(SIGINT, &sa, NULL))
   {
      logg_exit("failed sigaction for `%s': %s\n",
                  strsignal(SIGINT), strerror(errno));
   }

   /* Since we now have a connection, have it cleaned up on an orderly
      exit.
    */
   logg(3, "%s calling atexit()...\n", __func__);
   if (0 != atexit(finish_conn))
   {
      logg(0, "atexit failed on %s\n", strerror(errno));
   }
   logg(3, "%s done calling atexit().\n", __func__);

   /* Set the Postgres verbosity according to the logging verbosity.
    */
   PGVerbosity pg_verbosity;
   switch (opts.verbosity)
   {
   case 0:
      pg_verbosity = PQERRORS_TERSE;
      logg(3, "chose PGVerbosity terse.\n");
      break;
   case 1:
      pg_verbosity = PQERRORS_DEFAULT;
      logg(3, "chose PGVerbosity default.\n");
      break;
   default:
      pg_verbosity = PQERRORS_VERBOSE;
      logg(2, "chose PGVerbosity verbose.\n");
   }
   PQsetErrorVerbosity(conn, pg_verbosity);

   if (opts.run)
   {
      /* Set the mask for blocking SIG{INT,TERM} while a --run=CMD is in
         progress.
      */
      sigemptyset(&run_cmd_mask);
      sigaddset(&run_cmd_mask, SIGINT);
      sigaddset(&run_cmd_mask, SIGTERM);
      do_response = do_response_cmd;
   }
   else
   {
      do_response = do_response_sql;
   }

   logg(1, "connected to database `%s'.\n", PQdb(conn));
   return;
}
/*---------------------------------------------------------------------------*/

void
set_payload_flags ()
{
   int server_version = PQserverVersion(conn);
   // Payload is available from 9.0+ servers.
   is_pd_avail = (90000 <= server_version);
   if (opts.sql)
   {
      // We'll pass the payload to the SQL if it has '$1' in it.
      bool is_sql_parameterized = (bool) strstr(opts.sql, "$1");
      if (!is_pd_avail && is_sql_parameterized)
      {
         logg_exit("sql `%s' wants a payload but "
                   "postgres version %d cannot supply one.\n",
                   opts.sql, server_version);
      }
      is_pd_want = (is_pd_avail && is_sql_parameterized);
   }
   logg(3, "payload available: %d; wanted for SQL: %d\n",
        is_pd_avail, is_pd_want);
   return;
}
/*---------------------------------------------------------------------------*/

void
inner_loop ()
{
   logg(2, "acquiring socket from database connection...\n");
   int sock = PQsocket(conn);
   if (sock < 0)
   {
      logg(0, "error on PQsocket: %s\n", PQerrorMessage(conn));
      return;
   }

   set_payload_flags();

   /* Do the on-loop-entry response here, _after_ the LISTEN issued by
      listen_loop() but before waiting on the socket 'sock'.  This
      ordering aims to prevent the possibility of a gap between the
      on-loop-entry response and the LISTEN.  A NOTIFY during that gap
      would be missed because we would have finished the initial
      response but not yet asked PostgreSQL to send us NOTIFYs.
    */
   logg(2, "performing loop-entry response...\n");
   do_response(NULL);

   /* Create a file descriptor set containing the socket. */
   fd_set input_mask;
   FD_ZERO(&input_mask);
   FD_SET(sock, &input_mask);
   while (! shutdown_signalled)
   {
      logg(3, "waiting on select() inside the listen-forever loop...\n");
      if (select(FD_SETSIZE, &input_mask, NULL, NULL, NULL) < 0)
      {
         logg(0, "select() failed: %s\n", strerror(errno));
         break;
      }
      PQconsumeInput(conn);
      PGnotify *notify;
      while ((notify = PQnotifies(conn)) != NULL)
      {
         logg(3, "notify of '%s' received from pid %d\n",
                notify->relname, notify->be_pid);
         do_response(notify);
         PQfreemem(notify);
      }
   }
   logg(2, "falling out of %s.\n", __func__);
   return;
}
/*---------------------------------------------------------------------------*/

void
listen_loop()
{
   logg(1, "entering %s.\n", __func__);
   /* todo: maybe optionalize max recon attempts and sleep time.
    */
   const unsigned short rcn_attempt_max = 12;
   const unsigned short rcn_sleep_s = 5;
   /* Construct the LISTEN statement.
    */
   char *listen_sql;
   if (asprintf(&listen_sql, "LISTEN %s", opts.ntn_name) < 0)
   {
      logg_exit("failed to allocate listen_sql: %s\n", strerror(errno));
   }
   /* The number of reconnection attempts made since the connection was lost.
    */
   unsigned short rcn_attempt = 0;
   while (! shutdown_signalled)
   {
      /* Attempt to reconnect if we lose the database. */
      if (CONNECTION_OK != PQstatus(conn))
      {
         rcn_attempt++;
         if (rcn_attempt_max < rcn_attempt)
         {
            logg_exit("failed all %d reconnect attempts.\n", rcn_attempt_max);
         }
         /* PQreset() doesn't clear out the postgres error message, so
            use PQfinish() + PQconnectdb() instead.
          */
         PQfinish(conn);
         logg(0, "reconnect attempt %d in %ds...\n", rcn_attempt, rcn_sleep_s);
         sleep(rcn_sleep_s);
         conn = PQconnectdb(conninfo);
      }
      if (CONNECTION_OK == PQstatus(conn))
      {
         if (0 < rcn_attempt)
         {
            logg(1, "reconnect attempt %d succeeded.\n", rcn_attempt);
         }
         rcn_attempt = 0;
         logg(2, "about to issue listen command `%s'...\n", listen_sql);
         do_sql(listen_sql, NULL);
         inner_loop();
      }
      else
      {
         logg(0, "reconnect attempt %d failed: %s",
              rcn_attempt, PQerrorMessage(conn));
      }
   }
   free(listen_sql);
   if (NULL != opts.connect) free(conninfo);
   logg(2, "falling out of %s.\n", __func__);
   return;
}
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
